=== modified file 'database/schema/security.cfg'
--- database/schema/security.cfg	2011-07-14 10:05:11 +0000
+++ database/schema/security.cfg	2011-07-15 15:35:03 +0000
@@ -1058,6 +1058,7 @@
 public.sourcepackagepublishinghistory         = SELECT, INSERT
 public.sourcepackagerelease                   = SELECT
 public.sourcepackagereleasefile               = SELECT, INSERT, UPDATE
+public.teamparticipation                      = SELECT
 type=user
 
 [distroseriesdifferencejob]

=== modified file 'lib/canonical/launchpad/interfaces/_schema_circular_imports.py'
--- lib/canonical/launchpad/interfaces/_schema_circular_imports.py	2011-07-12 14:20:04 +0000
+++ lib/canonical/launchpad/interfaces/_schema_circular_imports.py	2011-07-15 14:28:21 +0000
@@ -404,6 +404,9 @@
 patch_entry_return_type(IArchive, 'newQueueAdmin', IArchivePermission)
 patch_plain_parameter_type(IArchive, 'syncSources', 'from_archive', IArchive)
 patch_plain_parameter_type(IArchive, 'syncSource', 'from_archive', IArchive)
+patch_plain_parameter_type(IArchive, 'copyPackage', 'from_archive', IArchive)
+patch_plain_parameter_type(
+    IArchive, 'copyPackages', 'from_archive', IArchive)
 patch_entry_return_type(IArchive, 'newSubscription', IArchiveSubscriber)
 patch_plain_parameter_type(
     IArchive, 'getArchiveDependency', 'dependency', IArchive)

=== modified file 'lib/lp/soyuz/browser/tests/test_archive_webservice.py'
--- lib/lp/soyuz/browser/tests/test_archive_webservice.py	2011-06-23 16:11:42 +0000
+++ lib/lp/soyuz/browser/tests/test_archive_webservice.py	2011-07-18 13:03:34 +0000
@@ -19,7 +19,12 @@
 from canonical.testing.layers import DatabaseFunctionalLayer
 from lp.app.interfaces.launchpad import ILaunchpadCelebrities
 from lp.registry.interfaces.pocket import PackagePublishingPocket
-from lp.soyuz.enums import ArchivePurpose
+from lp.services.features.testing import FeatureFixture
+from lp.soyuz.enums import (
+    ArchivePurpose,
+    PackagePublishingStatus,
+    )
+from lp.soyuz.interfaces.packagecopyjob import IPlainPackageCopyJobSource
 from lp.soyuz.interfaces.processor import IProcessorFamilySet
 from lp.testing import (
     celebrity_logged_in,
@@ -309,5 +314,67 @@
             [ws_arm], self.service.processor_families)
 
 
+class TestCopyPackage(WebServiceTestCase):
+    """Webservice test cases for the copyPackage/copyPackages methods"""
+
+    def setUp(self):
+        super(TestCopyPackage, self).setUp()
+        self.useFixture(FeatureFixture({u"soyuz.copypackage.enabled":  'on'}))
+
+    def setup_data(self):
+        self.ws_version = "devel"
+        uploader_dude = self.factory.makePerson()
+        source_archive = self.factory.makeArchive()
+        target_archive = self.factory.makeArchive(
+            purpose=ArchivePurpose.PRIMARY)
+        source = self.factory.makeSourcePackagePublishingHistory(
+            archive=source_archive, status=PackagePublishingStatus.PUBLISHED)
+        source_name = source.source_package_name
+        version = source.source_package_version
+        to_pocket = PackagePublishingPocket.RELEASE
+        to_series = self.factory.makeDistroSeries(
+            distribution=target_archive.distribution)
+        with person_logged_in(target_archive.owner):
+            target_archive.newComponentUploader(uploader_dude, "universe")
+        transaction.commit()
+        return (source_archive, source_name, target_archive, to_pocket,
+                to_series, uploader_dude, version)
+
+    def test_copyPackage(self):
+        """Basic smoke test"""
+        (source_archive, source_name, target_archive, to_pocket, to_series,
+         uploader_dude, version) = self.setup_data()
+
+        ws_target_archive = self.wsObject(target_archive, user=uploader_dude)
+        ws_source_archive = self.wsObject(source_archive)
+
+        ws_target_archive.copyPackage(
+            source_name=source_name, version=version,
+            from_archive=ws_source_archive, to_pocket=to_pocket.name,
+            to_series=to_series.name, include_binaries=False)
+        transaction.commit()
+
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        copy_job = job_source.getActiveJobs(target_archive).one()
+        self.assertEqual(target_archive, copy_job.target_archive)
+
+    def test_copyPackages(self):
+        """Basic smoke test"""
+        (source_archive, source_name, target_archive, to_pocket, to_series,
+         uploader_dude, version) = self.setup_data()
+
+        ws_target_archive = self.wsObject(target_archive, user=uploader_dude)
+        ws_source_archive = self.wsObject(source_archive)
+
+        ws_target_archive.copyPackages(
+            source_names=[source_name], from_archive=ws_source_archive,
+            to_pocket=to_pocket.name, to_series=to_series.name,
+            include_binaries=False)
+        transaction.commit()
+
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        copy_job = job_source.getActiveJobs(target_archive).one()
+        self.assertEqual(target_archive, copy_job.target_archive)
+
 def test_suite():
     return unittest.TestLoader().loadTestsFromName(__name__)

=== modified file 'lib/lp/soyuz/interfaces/archive.py'
--- lib/lp/soyuz/interfaces/archive.py	2011-06-24 21:21:36 +0000
+++ lib/lp/soyuz/interfaces/archive.py	2011-07-18 12:11:04 +0000
@@ -20,6 +20,7 @@
     'CannotUploadToArchive',
     'CannotUploadToPPA',
     'CannotUploadToPocket',
+    'ForbiddenByFeatureFlag',
     'FULL_COMPONENT_SUPPORT',
     'IArchive',
     'IArchiveAppend',
@@ -122,6 +123,12 @@
     """Exception raised when a copy cannot be performed."""
 
 
+@error_status(httplib.FORBIDDEN)
+class ForbiddenByFeatureFlag(Exception):
+    """Exception raised when using a method protected by a feature flag.
+    """
+
+
 @error_status(httplib.BAD_REQUEST)
 class CannotSwitchPrivacy(Exception):
     """Raised when switching the privacy of an archive that has
@@ -1218,6 +1225,91 @@
         :return: A new IArchiveAuthToken
         """
 
+    @call_with(person=REQUEST_USER)
+    @operation_parameters(
+        source_name=TextLine(title=_("Source package name")),
+        version=TextLine(title=_("Version")),
+        from_archive=Reference(schema=Interface),
+        # Really IArchive, see below
+        to_pocket=TextLine(title=_("Pocket name")),
+        to_series=TextLine(title=_("Distroseries name"), required=False),
+        include_binaries=Bool(
+            title=_("Include Binaries"),
+            description=_("Whether or not to copy binaries already built for"
+                          " this source"),
+            required=False))
+    @export_write_operation()
+    @operation_for_version('devel')
+    def copyPackage(source_name, version, from_archive, to_pocket,
+                    person, to_series=None, include_binaries=False):
+        """Copy a single named source into this archive.
+
+        Asynchronously copy a specific version of a named source to the
+        destination archive if necessary.  Calls to this method will return
+        immediately if the copy passes basic security checks and the copy
+        will happen sometime later with full checking.
+
+        :param source_name: a string name of the package to copy.
+        :param version: the version of the package to copy.
+        :param from_archive: the source archive from which to copy.
+        :param to_pocket: the target pocket (as a string).
+        :param to_series: the target distroseries (as a string).
+        :param include_binaries: optional boolean, controls whether or not
+            the published binaries for each given source should also be
+            copied along with the source.
+        :param person: the `IPerson` who requests the sync.
+
+        :raises NoSuchSourcePackageName: if the source name is invalid
+        :raises PocketNotFound: if the pocket name is invalid
+        :raises NoSuchDistroSeries: if the distro series name is invalid
+        :raises CannotCopy: if there is a problem copying.
+        """
+
+    @call_with(person=REQUEST_USER)
+    @operation_parameters(
+        source_names=List(
+            title=_("Source package names"),
+            value_type=TextLine()),
+        from_archive=Reference(schema=Interface),
+        #Really IArchive, see below
+        to_pocket=TextLine(title=_("Pocket name")),
+        to_series=TextLine(title=_("Distroseries name"), required=False),
+        include_binaries=Bool(
+            title=_("Include Binaries"),
+            description=_("Whether or not to copy binaries already built for"
+                          " this source"),
+            required=False))
+    @export_write_operation()
+    @operation_for_version('devel')
+    def copyPackages(source_names, from_archive, to_pocket, person,
+                     to_series=None, include_binaries=False):
+        """Atomically copy multiple named sources into this archive from another.
+
+        Asynchronously copy the most recent PUBLISHED versions of the named
+        sources to the destination archive if necessary.  Calls to this
+        method will return immediately if the copy passes basic security
+        checks and the copy will happen sometime later with full checking.
+
+        This operation will only succeed when all requested packages
+        are synchronised between the archives. If any of the requested
+        copies cannot be performed, the whole operation will fail. There
+        will be no partial changes of the destination archive.
+
+        :param source_names: a list of string names of packages to copy.
+        :param from_archive: the source archive from which to copy.
+        :param to_pocket: the target pocket (as a string).
+        :param to_series: the target distroseries (as a string).
+        :param include_binaries: optional boolean, controls whether or not
+            the published binaries for each given source should also be
+            copied along with the source.
+        :param person: the `IPerson` who requests the sync.
+
+        :raises NoSuchSourcePackageName: if the source name is invalid
+        :raises PocketNotFound: if the pocket name is invalid
+        :raises NoSuchDistroSeries: if the distro series name is invalid
+        :raises CannotCopy: if there is a problem copying.
+        """
+
 
 class IArchiveAppend(Interface):
     """Archive interface for operations restricted by append privilege."""

=== modified file 'lib/lp/soyuz/model/archive.py'
--- lib/lp/soyuz/model/archive.py	2011-06-24 21:21:36 +0000
+++ lib/lp/soyuz/model/archive.py	2011-07-18 12:59:05 +0000
@@ -102,6 +102,8 @@
 from lp.registry.interfaces.sourcepackagename import ISourcePackageNameSet
 from lp.registry.model.sourcepackagename import SourcePackageName
 from lp.registry.model.teammembership import TeamParticipation
+from lp.services.database.bulk import load_related
+from lp.services.features import getFeatureFlag
 from lp.services.job.interfaces.job import JobStatus
 from lp.services.propertycache import (
     cachedproperty,
@@ -115,6 +117,7 @@
     ArchivePurpose,
     ArchiveStatus,
     ArchiveSubscriberStatus,
+    PackageCopyPolicy,
     PackagePublishingStatus,
     PackageUploadStatus,
     )
@@ -130,6 +133,7 @@
     CannotUploadToPPA,
     ComponentNotFound,
     default_name_by_purpose,
+    ForbiddenByFeatureFlag,
     FULL_COMPONENT_SUPPORT,
     IArchive,
     IArchiveSet,
@@ -165,6 +169,7 @@
     IComponent,
     IComponentSet,
     )
+from lp.soyuz.interfaces.packagecopyjob import IPlainPackageCopyJobSource
 from lp.soyuz.interfaces.packagecopyrequest import IPackageCopyRequestSet
 from lp.soyuz.interfaces.processor import IProcessorFamilySet
 from lp.soyuz.interfaces.publishing import (
@@ -200,6 +205,7 @@
     )
 from lp.soyuz.model.section import Section
 from lp.soyuz.model.sourcepackagerelease import SourcePackageRelease
+from lp.soyuz.scripts.packagecopier import check_copy_permissions
 
 
 def storm_validate_external_dependencies(archive, attr, value):
@@ -1507,20 +1513,94 @@
             sources, to_pocket, to_series, include_binaries,
             person=person)
 
-    def syncSource(self, source_name, version, from_archive, to_pocket,
-                   to_series=None, include_binaries=False, person=None):
-        """See `IArchive`."""
+    def _validateAndFindSource(self, from_archive, source_name, version):
         # Check to see if the source package exists, and raise a useful error
         # if it doesn't.
         getUtility(ISourcePackageNameSet)[source_name]
         # Find and validate the source package version required.
         source = from_archive.getPublishedSources(
             name=source_name, version=version, exact_match=True).first()
+        return source
+
+    def syncSource(self, source_name, version, from_archive, to_pocket,
+                   to_series=None, include_binaries=False, person=None):
+        """See `IArchive`."""
+        source = self._validateAndFindSource(
+            from_archive, source_name, version)
 
         self._copySources(
             [source], to_pocket, to_series, include_binaries,
             person=person)
 
+    def copyPackage(self, source_name, version, from_archive, to_pocket,
+                    person, to_series=None, include_binaries=False):
+        """See `IArchive`."""
+        if not getFeatureFlag(u"soyuz.copypackage.enabled"):
+            raise ForbiddenByFeatureFlag
+
+        # Asynchronously copy a package using the job system.
+        pocket = self._text_to_pocket(to_pocket)
+        series = self._text_to_series(to_series)
+        # Upload permission checks, this will raise CannotCopy as
+        # necessary.
+        sourcepackagename = getUtility(ISourcePackageNameSet)[source_name]
+        check_copy_permissions(
+            person, self, series, pocket, [sourcepackagename])
+
+        self._validateAndFindSource(from_archive, source_name, version)
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        job_source.create(
+            package_name=source_name, source_archive=from_archive,
+            target_archive=self, target_distroseries=series,
+            target_pocket=pocket,
+            package_version=version, include_binaries=include_binaries,
+            copy_policy=PackageCopyPolicy.INSECURE, requester=person)
+
+    def copyPackages(self, source_names, from_archive, to_pocket,
+                     person, to_series=None, include_binaries=None):
+        """See `IArchive`."""
+        if not getFeatureFlag(u"soyuz.copypackage.enabled"):
+            raise ForbiddenByFeatureFlag
+
+        sources = self._collectLatestPublishedSources(
+            from_archive, source_names)
+        if not sources:
+            raise CannotCopy(
+                "None of the supplied package names are published")
+
+        # Bulk-load the sourcepackagereleases so that the list
+        # comprehension doesn't generate additional queries. The
+        # sourcepackagenames themselves will already have been loaded when
+        # generating the list of source publications in "sources".
+        load_related(
+            SourcePackageRelease, sources, ["sourcepackagereleaseID"])
+        sourcepackagenames = [source.sourcepackagerelease.sourcepackagename
+                              for source in sources]
+
+        # Now do a mass check of permissions.
+        pocket = self._text_to_pocket(to_pocket)
+        series = self._text_to_series(to_series)
+        check_copy_permissions(
+            person, self, series, pocket, sourcepackagenames)
+
+        # If we get this far then we can create the PackageCopyJob.
+        copy_tasks = []
+        for source in sources:
+            task = (
+                source.sourcepackagerelease.sourcepackagename,
+                source.sourcepackagerelease.version,
+                from_archive,
+                self,
+                PackagePublishingPocket.RELEASE
+                )
+            copy_tasks.append(task)
+
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        job_source.createMultiple(
+            series, copy_tasks, person,
+            copy_policy=PackageCopyPolicy.INSECURE,
+            include_binaries=include_binaries)
+
     def _collectLatestPublishedSources(self, from_archive, source_names):
         """Private helper to collect the latest published sources for an
         archive.
@@ -1528,6 +1608,9 @@
         :raises NoSuchSourcePackageName: If any of the source_names do not
             exist.
         """
+        # XXX bigjools bug=810421
+        # This code is inefficient.  It should try to bulk load all the
+        # sourcepackagenames and publications instead of iterating.
         sources = []
         for name in source_names:
             # Check to see if the source package exists. This will raise
@@ -1544,6 +1627,27 @@
                 sources.append(first_source)
         return sources
 
+    def _text_to_series(self, to_series):
+        if to_series is not None:
+            result = getUtility(IDistroSeriesSet).queryByName(
+                self.distribution, to_series)
+            if result is None:
+                raise NoSuchDistroSeries(to_series)
+            series = result
+        else:
+            series = None
+
+        return series
+
+    def _text_to_pocket(self, to_pocket):
+        # Convert the to_pocket string to its enum.
+        try:
+            pocket = PackagePublishingPocket.items[to_pocket.upper()]
+        except KeyError:
+            raise PocketNotFound(to_pocket.upper())
+
+        return pocket
+
     def _copySources(self, sources, to_pocket, to_series=None,
                      include_binaries=False, person=None):
         """Private helper function to copy sources to this archive.
@@ -1553,12 +1657,8 @@
         """
         # Circular imports.
         from lp.soyuz.scripts.packagecopier import do_copy
-        # Convert the to_pocket string to its enum.
-        try:
-            pocket = PackagePublishingPocket.items[to_pocket.upper()]
-        except KeyError:
-            raise PocketNotFound(to_pocket.upper())
 
+        pocket = self._text_to_pocket(to_pocket)
         # Fail immediately if the destination pocket is not Release and
         # this archive is a PPA.
         if self.is_ppa and pocket != PackagePublishingPocket.RELEASE:
@@ -1566,14 +1666,7 @@
                 "Destination pocket must be 'release' for a PPA.")
 
         # Now convert the to_series string to a real distroseries.
-        if to_series is not None:
-            result = getUtility(IDistroSeriesSet).queryByName(
-                self.distribution, to_series)
-            if result is None:
-                raise NoSuchDistroSeries(to_series)
-            series = result
-        else:
-            series = None
+        series = self._text_to_series(to_series)
 
         # Perform the copy, may raise CannotCopy. Don't do any further
         # permission checking: this method is protected by

=== modified file 'lib/lp/soyuz/model/packagecopyjob.py'
--- lib/lp/soyuz/model/packagecopyjob.py	2011-07-15 12:09:45 +0000
+++ lib/lp/soyuz/model/packagecopyjob.py	2011-07-15 15:35:03 +0000
@@ -482,13 +482,14 @@
         do_copy(
             sources=[source_package], archive=self.target_archive,
             series=self.target_distroseries, pocket=self.target_pocket,
-            include_binaries=self.include_binaries, check_permissions=False,
-            overrides=[override], send_email=send_email)
+            include_binaries=self.include_binaries, check_permissions=True,
+            person=self.requester, overrides=[override],
+            send_email=send_email)
 
         if pu is not None:
             # A PackageUpload will only exist if the copy job had to be
             # held in the queue because of policy/ancestry checks.  If one
-            # does exist we need to make sure.
+            # does exist we need to make sure it gets moved to DONE.
             pu.setDone()
 
     def abort(self):

=== modified file 'lib/lp/soyuz/tests/test_archive.py'
--- lib/lp/soyuz/tests/test_archive.py	2011-06-24 09:53:18 +0000
+++ lib/lp/soyuz/tests/test_archive.py	2011-07-18 12:50:15 +0000
@@ -39,6 +39,7 @@
     )
 from lp.registry.interfaces.pocket import PackagePublishingPocket
 from lp.registry.interfaces.series import SeriesStatus
+from lp.services.features.testing import FeatureFixture
 from lp.services.job.interfaces.job import JobStatus
 from lp.services.propertycache import clear_property_cache
 from lp.services.worlddata.interfaces.country import ICountrySet
@@ -49,14 +50,17 @@
     ArchivePermissionType,
     ArchivePurpose,
     ArchiveStatus,
+    PackageCopyPolicy,
     PackagePublishingStatus,
     )
 from lp.soyuz.interfaces.archive import (
     ArchiveDependencyError,
     ArchiveDisabled,
+    CannotCopy,
     CannotRestrictArchitectures,
     CannotUploadToPocket,
     CannotUploadToPPA,
+    ForbiddenByFeatureFlag,
     IArchiveSet,
     InsufficientUploadRights,
     InvalidPocketForPartnerArchive,
@@ -69,6 +73,7 @@
 from lp.soyuz.interfaces.archivepermission import IArchivePermissionSet
 from lp.soyuz.interfaces.binarypackagename import IBinaryPackageNameSet
 from lp.soyuz.interfaces.component import IComponentSet
+from lp.soyuz.interfaces.packagecopyjob import IPlainPackageCopyJobSource
 from lp.soyuz.interfaces.processor import IProcessorFamilySet
 from lp.soyuz.model.archive import Archive
 from lp.soyuz.model.archivepermission import ArchivePermission
@@ -1933,10 +1938,37 @@
             filtered_sources])
 
 
+class TestSyncSourceFeatureFlag(TestCaseWithFactory):
+
+    layer = DatabaseFunctionalLayer
+
+    def test_copyPackage_requires_feature_flag(self):
+        # Ensure feature is off.
+        self.useFixture(FeatureFixture({u"soyuz.copypackage.enabled": ''}))
+        archive = self.factory.makeArchive()
+        self.assertRaises(
+            ForbiddenByFeatureFlag,
+            archive.copyPackage,
+            None, None, None, None, None)
+
+    def test_copyPackages_requires_feature_flag(self):
+        # Ensure feature is off.
+        self.useFixture(FeatureFixture({u"soyuz.copypackage.enabled": ''}))
+        archive = self.factory.makeArchive()
+        self.assertRaises(
+            ForbiddenByFeatureFlag,
+            archive.copyPackages,
+            None, None, None, None, None)
+
+
 class TestSyncSource(TestCaseWithFactory):
 
     layer = DatabaseFunctionalLayer
 
+    def setUp(self):
+        super(TestSyncSource, self).setUp()
+        self.useFixture(FeatureFixture({u"soyuz.copypackage.enabled": 'on'}))
+
     def test_security_team_can_copy_to_primary(self):
         # A member of ubuntu-security can use syncSource on any package
         # in the Ubuntu primary archive, regardless of their normal
@@ -1965,3 +1997,190 @@
             1,
             ubuntu.main_archive.getPublishedSources(
                 name=source.source_package_name).count())
+
+    def _setup_copy_data(self, target_purpose=None):
+        if target_purpose is None:
+            target_purpose = ArchivePurpose.PPA
+        source_archive = self.factory.makeArchive()
+        target_archive = self.factory.makeArchive(purpose=target_purpose)
+        source = self.factory.makeSourcePackagePublishingHistory(
+            archive=source_archive, status=PackagePublishingStatus.PUBLISHED)
+        source_name = source.source_package_name
+        version = source.source_package_version
+        to_pocket = PackagePublishingPocket.RELEASE
+        to_series = self.factory.makeDistroSeries(
+            distribution=target_archive.distribution)
+        return (source, source_archive, source_name, target_archive,
+                to_pocket, to_series, version)
+
+    def test_copyPackage_creates_packagecopyjob(self):
+        # The copyPackage method should create a PCJ with the appropriate
+        # parameters.
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data()
+        with person_logged_in(target_archive.owner):
+            target_archive.copyPackage(
+                source_name, version, source_archive, to_pocket.name,
+                to_series=to_series.name, include_binaries=False,
+                person=target_archive.owner)
+
+        # The source should not be published yet in the target_archive.
+        published = target_archive.getPublishedSources(
+            name=source.source_package_name).any()
+        self.assertIs(None, published)
+
+        # There should be one copy job.
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        copy_job = job_source.getActiveJobs(target_archive).one()
+
+        # Its data should reflect the requested copy.
+        self.assertEqual(source_name, copy_job.package_name)
+        self.assertEqual(version, copy_job.package_version)
+        self.assertEqual(target_archive, copy_job.target_archive)
+        self.assertEqual(source_archive, copy_job.source_archive)
+        self.assertEqual(to_series, copy_job.target_distroseries)
+        self.assertEqual(to_pocket, copy_job.target_pocket)
+        self.assertFalse(copy_job.include_binaries)
+        self.assertEquals(PackageCopyPolicy.INSECURE, copy_job.copy_policy)
+
+    def test_copyPackage_disallows_non_primary_archive_uploaders(self):
+        # If copying to a primary archive and you're not an uploader for
+        # the package then you can't copy.
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data(
+            target_purpose=ArchivePurpose.PRIMARY)
+        person = self.factory.makePerson()
+        self.assertRaises(
+            CannotCopy,
+            target_archive.copyPackage, source_name, version, source_archive,
+            to_pocket.name, to_series=to_series.name, include_binaries=False,
+            person=person)
+
+    def test_copyPackage_allows_primary_archive_uploaders(self):
+        # Copying to a primary archive if you're already an uploader is OK.
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data(
+            target_purpose=ArchivePurpose.PRIMARY)
+        person = self.factory.makePerson()
+        with person_logged_in(target_archive.owner):
+            target_archive.newComponentUploader(person, "universe")
+        target_archive.copyPackage(
+            source_name, version, source_archive, to_pocket.name,
+            to_series=to_series.name, include_binaries=False,
+            person=person)
+
+        # There should be one copy job.
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        copy_job = job_source.getActiveJobs(target_archive).one()
+        self.assertEqual(target_archive, copy_job.target_archive)
+
+    def test_copyPackage_disallows_non_PPA_owners(self):
+        # Only people with launchpad.Append are allowed to call copyPackage.
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data()
+        person = self.factory.makePerson()
+        self.assertTrue(target_archive.is_ppa)
+        self.assertRaises(
+            CannotCopy,
+            target_archive.copyPackage, source_name, version, source_archive,
+            to_pocket.name, to_series=to_series.name, include_binaries=False,
+            person=person)
+
+    def test_copyPackage_disallows_non_release_target_pocket_for_PPA(self):
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data()
+        to_pocket = PackagePublishingPocket.UPDATES
+        self.assertTrue(target_archive.is_ppa)
+        self.assertRaises(
+            CannotCopy,
+            target_archive.copyPackage, source_name, version, source_archive,
+            to_pocket.name, to_series=to_series.name, include_binaries=False,
+            person=target_archive.owner)
+
+    def test_copyPackages_with_single_package(self):
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data()
+
+        with person_logged_in(target_archive.owner):
+            target_archive.copyPackages(
+                [source_name], source_archive, to_pocket.name,
+                to_series=to_series.name, include_binaries=False,
+                person=target_archive.owner)
+
+        # The source should not be published yet in the target_archive.
+        published = target_archive.getPublishedSources(
+            name=source.source_package_name).any()
+        self.assertIs(None, published)
+
+        # There should be one copy job.
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        copy_job = job_source.getActiveJobs(target_archive).one()
+        self.assertEqual(target_archive, copy_job.target_archive)
+
+    def test_copyPackages_with_multiple_packages(self):
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data()
+        sources = [source]
+        sources.append(self.factory.makeSourcePackagePublishingHistory(
+            archive=source_archive,
+            status=PackagePublishingStatus.PUBLISHED))
+        sources.append(self.factory.makeSourcePackagePublishingHistory(
+            archive=source_archive,
+            status=PackagePublishingStatus.PUBLISHED))
+        names = [source.sourcepackagerelease.sourcepackagename.name
+                 for source in sources]
+
+        with person_logged_in(target_archive.owner):
+            target_archive.copyPackages(
+                names, source_archive, to_pocket.name,
+                to_series=to_series.name, include_binaries=False,
+                person=target_archive.owner)
+
+        # Make sure three copy jobs exist.
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        copy_jobs = job_source.getActiveJobs(target_archive)
+        self.assertEqual(3, copy_jobs.count())
+
+    def test_copyPackages_disallows_non_primary_archive_uploaders(self):
+        # If copying to a primary archive and you're not an uploader for
+        # the package then you can't copy.
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data(
+            target_purpose=ArchivePurpose.PRIMARY)
+        person = self.factory.makePerson()
+        self.assertRaises(
+            CannotCopy,
+            target_archive.copyPackages, [source_name], source_archive,
+            to_pocket.name, to_series=to_series.name, include_binaries=False,
+            person=person)
+
+    def test_copyPackages_allows_primary_archive_uploaders(self):
+        # Copying to a primary archive if you're already an uploader is OK.
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data(
+            target_purpose=ArchivePurpose.PRIMARY)
+        person = self.factory.makePerson()
+        with person_logged_in(target_archive.owner):
+            target_archive.newComponentUploader(person, "universe")
+        target_archive.copyPackages(
+            [source_name], source_archive, to_pocket.name,
+            to_series=to_series.name, include_binaries=False,
+            person=person)
+
+        # There should be one copy job.
+        job_source = getUtility(IPlainPackageCopyJobSource)
+        copy_job = job_source.getActiveJobs(target_archive).one()
+        self.assertEqual(target_archive, copy_job.target_archive)
+
+    def test_copyPackages_disallows_non_PPA_owners(self):
+        # Only people with launchpad.Append are allowed to call copyPackage.
+        (source, source_archive, source_name, target_archive, to_pocket,
+         to_series, version) = self._setup_copy_data()
+        person = self.factory.makePerson()
+        self.assertTrue(target_archive.is_ppa)
+        self.assertRaises(
+            CannotCopy,
+            target_archive.copyPackages, [source_name], source_archive,
+            to_pocket.name, to_series=to_series.name, include_binaries=False,
+            person=person)
+

=== modified file 'lib/lp/soyuz/tests/test_packagecopyjob.py'
--- lib/lp/soyuz/tests/test_packagecopyjob.py	2011-07-15 12:09:45 +0000
+++ lib/lp/soyuz/tests/test_packagecopyjob.py	2011-07-15 16:59:08 +0000
@@ -3,6 +3,7 @@
 
 """Tests for sync package jobs."""
 
+import operator
 from testtools.content import text_content
 from testtools.matchers import (
     Equals,
@@ -333,6 +334,8 @@
 
         source = getUtility(IPlainPackageCopyJobSource)
         requester = self.factory.makePerson()
+        with person_logged_in(target_archive.owner):
+            target_archive.newComponentUploader(requester, "main")
         job = source.create(
             package_name="libc",
             source_archive=breezy_archive, target_archive=target_archive,
@@ -406,6 +409,8 @@
             target_pocket=PackagePublishingPocket.RELEASE,
             package_version="2.8-1", include_binaries=False,
             requester=requester)
+        with person_logged_in(archive2.owner):
+            archive2.newComponentUploader(requester, "main")
         transaction.commit()
 
         out, err, exit_code = run_script(
@@ -545,6 +550,8 @@
 
         source = getUtility(IPlainPackageCopyJobSource)
         requester = self.factory.makePerson()
+        with person_logged_in(target_archive.owner):
+            target_archive.newComponentUploader(requester, "restricted")
         job = source.create(
             package_name="libc",
             package_version="2.8-1",
@@ -584,6 +591,8 @@
         # there's no ancestry.
         source = getUtility(IPlainPackageCopyJobSource)
         requester = self.factory.makePerson()
+        with person_logged_in(target_archive.owner):
+            target_archive.newComponentUploader(requester, "main")
         job = source.create(
             package_name="copyme",
             package_version="2.8-1",
@@ -736,6 +745,7 @@
         publisher = SoyuzTestPublisher()
         publisher.prepareBreezyAutotest()
         distroseries = publisher.breezy_autotest
+        distroseries.changeslist = "changes@example.com"
 
         target_archive = self.factory.makeArchive(
             distroseries.distribution, purpose=ArchivePurpose.PRIMARY)
@@ -748,7 +758,9 @@
             archive=source_archive)
 
         source = getUtility(IPlainPackageCopyJobSource)
-        requester = self.factory.makePerson()
+        requester = self.factory.makePerson(email="requester@example.com")
+        with person_logged_in(target_archive.owner):
+            target_archive.newComponentUploader(requester, "main")
         job = source.create(
             package_name="copyme",
             package_version="2.8-1",
@@ -771,6 +783,9 @@
         pu = getUtility(IPackageUploadSet).getByPackageCopyJobIDs(
             [removeSecurityProxy(job).context.id]).one()
         pu.acceptFromQueue()
+        # Clear existing emails so we can see only the ones the job
+        # generates later.
+        pop_notifications()
         self.runJob(job)
 
         # The job should have set the PU status to DONE:
@@ -780,6 +795,16 @@
         existing_sources = target_archive.getPublishedSources(name='copyme')
         self.assertIsNot(None, existing_sources.any())
 
+        # It would be nice to test emails in a separate test but it would
+        # require all of the same setup as above again so we might as well
+        # do it here.
+        emails = pop_notifications(sort_key=operator.itemgetter('To'))
+
+        # We expect an uploader email and an announcement to the changeslist.
+        self.assertEquals(2, len(emails))
+        self.assertIn("requester@example.com", emails[0]['To'])
+        self.assertIn("changes@example.com", emails[1]['To'])
+
     def test_findMatchingDSDs_matches_all_DSDs_for_job(self):
         # findMatchingDSDs finds matching DSDs for any of the packages
         # in the job.

